import threading
import time

from termcolor import colored
from colorama import init as colorama_init, Fore, Style

from dweet2ser.settings import timestamp, internet_connection, s_print

colorama_init()


def _print_device_list(dev_list):
    """
    Prints a list of devices in table form.
    """
    cols = ["#".ljust(3),
            "Name".ljust(16),
            "Type".ljust(10),
            "Port".ljust(15),
            "ThingName".ljust(20),
            "Locked".ljust(10),
            "Muted".ljust(10)
            ]
    header = ''
    for col in cols:
        header = header + f"{col}  "
    print(f"\t{Fore.LIGHTWHITE_EX}{header}{Style.RESET_ALL}")

    for i in range(0, len(dev_list)):
        num = str(i + 1)
        d = dev_list[i]
        if type(d).__name__ == "LocalDevice":
            print(f"\t"
                  f"{num.ljust(3)}  "
                  f"{d.name.ljust(16)}  "
                  f"{colored(d.type.ljust(10), d.type_color)}  "
                  f"{d.port_name.ljust(15)}  "
                  f"{''.ljust(20)}  "
                  f"{''.ljust(10)}  "
                  f"{str(d.mute).ljust(10)}  "
                  )
        if type(d).__name__ == "RemoteDevice":
            print(f"\t"
                  f"{num.ljust(3)}  "
                  f"{d.name.ljust(16)}  "
                  f"{colored(d.type.ljust(10), d.type_color)}  "
                  f"{''.ljust(15)}  "
                  f"{d.thing_id.ljust(20)}  "
                  f"{str(d.locked).ljust(10)}  "
                  f"{str(d.mute).ljust(10)}  "
                  )


class DeviceBus(object):
    """
    Holds all the devices in a connection and facilitates communication between them. All DCE devices
    write to all DTE devices and vice versa.
    """
    def __init__(self):
        self.dce_devices = []
        self.dte_devices = []
        self.listen_threads = {}

        self.stream_restarter = threading.Thread(target=self._check_for_crashed_threads)
        self.stream_restarter.daemon = True
        self.stream_restarter.start()

    def add_device(self, device):
        """
        Adds the given device to the connection, and starts a listening thread if the device is not muted.
        """
        if device.mode == "DTE":
            self.dte_devices.append(device)
        if device.mode == "DCE":
            self.dce_devices.append(device)
        if not device.mute:
            self.listen_threads[device.name] = threading.Thread(target=self._listen_stream, args=[device])
            self.listen_threads[device.name].daemon = True
            self.listen_threads[device.name].start()
        return True

    def remove_device(self, device_name):
        """
        Takes a device name, and removes that device if the name is found
        """
        found = False
        for d in self.dce_devices:
            if d.name == device_name:
                found = True
                d.kill_listen_stream()
                self.dce_devices.remove(d)
                print(f"{timestamp()}Device '{d.name}' removed.")

        for d in self.dte_devices:
            if d.name == device_name:
                found = True
                d.kill_listen_stream()
                self.dte_devices.remove(d)
                print(f"{timestamp()}Device '{d.name}' removed.")

        if not found:
            print("Device not found.")

    def print_status(self):
        """
        Prints a list of all devices in the connection.
        """
        print("\nDCE Devices")
        _print_device_list(self.dce_devices)
        print("\nDTE Devices")
        _print_device_list(self.dte_devices)

    def print_threads(self):
        print(self.listen_threads)

    def _listen_stream(self, device):
        """
        Listens for messages from the given device, then writes to all appropriate devices on bus.
        """
        s_print(f"{timestamp()}Listen stream started for {device.name}.")

        for message in device.listen():
            message = str(message)
            message_decoded = bytes.fromhex(message).decode('latin-1').rstrip().replace('\n', '<NL>').replace('\r', '')

            s_print(f"\n{timestamp()}Received {colored(device.type, device.type_color)} message from {device.name}:"
                    f" {Fore.LIGHTWHITE_EX}{message_decoded}{Style.RESET_ALL}")

            if device.mode == "DTE":
                # Messages from DTE devices get sent to all DCE devices.
                for d in self.dce_devices:
                    d.write(message)

            elif device.mode == "DCE":
                # Messages from DCE devices get sent to all DTE devices.
                for d in self.dte_devices:
                    d.write(message)

            else:
                s_print("Mode not found")

        return True

    def _check_for_crashed_threads(self):
        """
        If a thread has thrown an error, attempt to restart it.
        """
        while True:
            for d in self.dce_devices:
                if d.exc and internet_connection():
                    # if the thread had an exception, start a new one
                    self._restart_thread(d)
            for d in self.dte_devices:
                if d.exc and internet_connection():
                    # if the thread had an exception, start a new one
                    self._restart_thread(d)
            time.sleep(.01)

    def _restart_thread(self, device):
        s_print(f"{timestamp()}Reconnecting to {device.name}.")
        device.kill_listen_stream()  # attempt to kill the listen thread if it is still responding
        device.exc = False  # reset the exception flag to false
        device.restart_session()  # start a new Session
        device.send_message_queue()  # send any messages saved in the queue
        if not device.mute:  # restart the listen thread if needed
            self.listen_threads[device.name] = threading.Thread(target=self._listen_stream, args=[device])
            self.listen_threads[device.name].daemon = True
            self.listen_threads[device.name].start()
        return



import os
from configparser import ConfigParser

from dweet2ser.local_device import LocalDevice
from dweet2ser.remote_device import RemoteDevice
from dweet2ser.settings import sys_stamp, CONFIG_DEFAULTS, DEFAULT_CONFIG_FILE, s_print


class Dweet2serConfiguration(object):

    def __init__(self, bus):
        self.parser = ConfigParser()
        self.bus = bus

    def add_devices_from_file(self, file=DEFAULT_CONFIG_FILE):
        """ Attempt to add devices from an arbitrary file. If no file is given the default file is used.

        """
        if self._load_config_file(file):
            try:
                self._add_devices()
            except Exception as exc:
                s_print(f"{sys_stamp}Invalid config file format: {exc}")

    def _add_devices(self):
        devices = self.parser.sections()
        # make sure there is actually something in config
        if len(devices) > 0:
            s_print(f"{sys_stamp}Loading devices from config.")
            for d in devices:
                name = d
                location = self.parser[d]["location"]
                dev_type = self.parser[d]["type"]
                if self.parser[d]["mute"].upper().strip() == "TRUE":
                    mute = True
                else:
                    mute = False
                if location == "local":
                    port = self.parser[d]["port"]
                    try:
                        dev = LocalDevice(port, dev_type, name, mute)
                        self.bus.add_device(dev)
                        s_print(f"{sys_stamp}Added {location} {dev_type} device '{name}' from config.")
                    except Exception as e:
                        s_print(f"{sys_stamp}Failed to add device '{name}' from default config: {e}")
                elif location == "remote":
                    thing_name = self.parser[d]["thing_name"]
                    if self.parser[d]["key"] == "" or self.parser[d]["key"].lower().strip() == "none":
                        key = None
                    else:
                        key = self.parser[d]["key"]
                    try:
                        dev = RemoteDevice(thing_name, dev_type, key, name, mute)
                        self.bus.add_device(dev)
                        s_print(f"{sys_stamp}Added {location} {dev_type} device '{name}' from config.")
                    except Exception as e:
                        s_print(f"{sys_stamp}Failed to add device '{name}' from default config: {e}")
                else:
                    s_print(f"{sys_stamp}Failed to add device '{name}' from default config: "
                            f"invalid location '{location}'")

    def save_current_to_file(self):
        """ Save the current device bus to default config file

        """
        self.parser = ConfigParser()  # clear any settings in the parser
        self._add_defaults_to_parser()

        # add DCE settings to parser
        for dev in self.bus.dce_devices:
            self.parser.add_section(dev.name)
            self.parser[dev.name]["type"] = "DCE"
            if type(dev).__name__ == "LocalDevice":
                self.parser[dev.name]["location"] = "local"
                self.parser[dev.name]["port"] = dev.port_name
            if type(dev).__name__ == "RemoteDevice":
                self.parser[dev.name]["location"] = "remote"
                self.parser[dev.name]["thing_name"] = dev.thing_id
                self.parser[dev.name]["key"] = str(dev.thing_key)
            self.parser[dev.name]["mute"] = str(dev.mute)

        # add DTE settings to parser
        for dev in self.bus.dte_devices:
            self.parser.add_section(dev.name)
            self.parser[dev.name]["type"] = "DTE"
            if type(dev).__name__ == "LocalDevice":
                self.parser[dev.name]["location"] = "local"
                self.parser[dev.name]["port"] = dev.port_name
            if type(dev).__name__ == "RemoteDevice":
                self.parser[dev.name]["location"] = "remote"
                self.parser[dev.name]["thing_name"] = dev.thing_id
                self.parser[dev.name]["key"] = str(dev.thing_key)
            self.parser[dev.name]["mute"] = str(dev.mute)

        os.makedirs(os.path.dirname(DEFAULT_CONFIG_FILE), exist_ok=True)

        with open(DEFAULT_CONFIG_FILE, 'w') as configfile:
            configfile.write("\n; The settings in the [DEFAULT] section reference the values set in 'settings.py' "
                             "in the package directory."
                             "\n; They should be here for config file stability."
                             "\n; If you would like to change these default settings permanently, "
                             "you should do so in settings.py, not here."
                             "\n; Make sure you know what you're doing.\n\n")
        with open(DEFAULT_CONFIG_FILE, 'a') as configfile:
            self.parser.write(configfile)

        s_print(f"{sys_stamp}Settings saved to config file: {DEFAULT_CONFIG_FILE}")

    def _load_config_file(self, file):
        if not os.path.isabs(file):
            file = os.path.join(os.getcwd(), file)
        if os.path.exists(file):
            s_print(f"{sys_stamp}Found config file at {file}.")
            try:
                self.parser = ConfigParser()
                self.parser.read(file)
                return True
            except Exception as e:
                s_print(f"{sys_stamp}Failed to read config file: {e}")
                return False

        else:
            self._add_defaults_to_parser()
            s_print(f"{sys_stamp}Config file {file} not found.")
            if file == DEFAULT_CONFIG_FILE:
                s_print(f"{sys_stamp}Creating empty default file at: {file}")
                os.makedirs(os.path.dirname(DEFAULT_CONFIG_FILE), exist_ok=True)
                with open(DEFAULT_CONFIG_FILE, 'w') as configfile:
                    self.parser.write(configfile)
            return False

    def _add_defaults_to_parser(self):
        self.parser["DEFAULT"] = CONFIG_DEFAULTS



# -*- coding: utf-8 -*-

# future imports
from __future__ import absolute_import
from __future__ import unicode_literals

# local imports
from .api import DweepyError
from .api import dweet
from .api import dweet_for
from .api import get_alert
from .api import get_dweets_for
from .api import get_latest_dweet_for
from .api import lock
from .api import remove_alert
from .api import remove_lock
from .api import set_alert
from .api import unlock
from .streaming import listen_for_dweets_from


# all of the following objects will be imported if the caller does
# `from dweepy import *` (don't)
__all__ = [
    'DweepyError', 'dweet', 'dweet_for', 'get_alert', 'get_dweets_for',
    'get_latest_dweet_for', 'listen_for_dweets_from', 'lock', 'remove_alert',
    'remove_lock', 'set_alert', 'unlock',
]




import time

import serial
from termcolor import colored

from dweet2ser.settings import timestamp, s_print


class LocalDevice(object):
    """
    A device connected to a serial port on the local machine.
    """
    def __init__(self, port, mode, name="Local Device", mute=False):
        self.name = name
        self.type = "serial"
        self.type_color = "red"
        self.port_name = port
        self.mode = mode
        self.serial_port = serial.Serial(self.port_name)
        self._last_message = ''
        self.mute = mute
        self.exc = False
        self.listening = False

    def write(self, message: str):
        """
        Receives a hex string, converts to bytes and writes to the serial port.
        """
        if type(message) is not str:  # make sure the message is a string
            message = str(message)
        message_bytes = bytes.fromhex(message)  # convert dweet string into bytes for RS232.
        message_decoded = message_bytes.decode('latin-1').rstrip()
        s_print(f"{timestamp()}{colored(self.type.capitalize(), self.type_color)} message sent to {self.name}: {message_decoded}")
        return self.serial_port.write(message_bytes)

    def listen(self):
        """listens to serial port, yields what it hears
        """
        ser = self.serial_port
        self.listening = True

        # TODO: test with a variety of devices and protocols
        while self.listening:
            if ser.in_waiting > 0:
                time.sleep(.1) # wait until all the data is in the buffer
                ser_data = ser.read(ser.in_waiting)
                self._last_message = ser_data.hex()
                yield ser_data.hex()
            else:
                time.sleep(0.0001)

    def kill_listen_stream(self):
        """
        Stops any threads listening to the serial port.
        """
        self.listening = False

    def get_last_message(self):
        """
        Returns the last message read from the serial port.
        """
        return self._last_message



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# marty mcfly imports
from __future__ import absolute_import
from __future__ import unicode_literals

# stdlib imports
import datetime
import json

# third-party imports
import requests
from requests.exceptions import ChunkedEncodingError


# python 2/3 compatibility shim for checking if value is a text type
try:
    basestring  # attempt to evaluate basestring

    def isstr(s):
        return isinstance(s, basestring)
except NameError:
    def isstr(s):
        return isinstance(s, str)


# base url for all requests
BASE_URL = 'https://dweet.io'


def _check_stream_timeout(started, timeout):
    """Check if the timeout has been reached and raise a `StopIteration` if so.
    """
    if timeout:
        elapsed = datetime.datetime.utcnow() - started
        if elapsed.seconds > timeout:
            raise StopIteration


def _listen_for_dweets_from_response(response):
    """Yields dweets as received from dweet.io's streaming API
    """
    streambuffer = ''
    for byte in response.iter_content(chunk_size=2000):
        if byte:
            streambuffer += byte.decode('ascii')
            try:
                dweet = json.loads(streambuffer.splitlines()[1])
            except (IndexError, ValueError):
                continue
            if isstr(dweet):
                yield json.loads(dweet)
            streambuffer = ''


def listen_for_dweets_from(thing_name, timeout=900, key=None, session=None):
    """Create a real-time subscription to dweets
    """
    url = BASE_URL + '/listen/for/dweets/from/{0}'.format(thing_name)
    session = session or requests.Session()
    if key is not None:
        params = {'key': key}
    else:
        params = None

    start = datetime.datetime.utcnow()
    while True:
        request = requests.Request("GET", url, params=params).prepare()
        resp = session.send(request, stream=True, timeout=timeout)
        try:
            for x in _listen_for_dweets_from_response(resp):
                yield x
                _check_stream_timeout(start, timeout)
        except (ChunkedEncodingError, requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout):
            print("Error in listen_for_dweets_from")  # test message
            pass
        _check_stream_timeout(start, timeout)



# -*- coding: utf-8 -*-

# future imports
from __future__ import absolute_import
from __future__ import unicode_literals

# stdlib imports
import json

try:
    # python 3
    from urllib.parse import quote
except ImportError:
    # python 2
    from urllib import quote

# third-party imports
import requests

# base url for all requests
BASE_URL = 'https://dweet.io'


class DweepyError(Exception):
    pass


def _request(method, url, session=None, **kwargs):
    """Make HTTP request, raising an exception if it fails.
    """
    url = BASE_URL + url

    if session:
        request_func = getattr(session, method)
    else:
        request_func = getattr(requests, method)
    response = request_func(url, **kwargs)
    # raise an exception if request is not successful
    if not response.status_code == requests.codes.ok:
        raise DweepyError('HTTP {0} response'.format(response.status_code))
    response_json = response.json()
    if response_json['this'] == 'failed':
        raise DweepyError(response_json['because'])
    return response_json['with']


def _send_dweet(payload, url, params=None, session=None):
    """Send a dweet to dweet.io
    """
    data = json.dumps(payload)
    headers = {'Content-type': 'application/json'}
    return _request('post', url, data=data, headers=headers, params=params, session=session)


def dweet(payload, session=None):
    """Send a dweet to dweet.io without naming your thing
    """
    return _send_dweet(payload, '/dweet', session=session)


def dweet_for(thing_name, payload, key=None, session=None):
    """Send a dweet to dweet.io for a thing with a known name
    """
    if key is not None:
        params = {'key': key}
    else:
        params = None
    return _send_dweet(payload, '/dweet/for/{0}'.format(thing_name), params=params, session=session)


def get_latest_dweet_for(thing_name, key=None, session=None):
    """Read the latest dweet for a dweeter
    """
    if key is not None:
        params = {'key': key}
    else:
        params = None
    return _request('get', '/get/latest/dweet/for/{0}'.format(thing_name), params=params, session=session)


def get_dweets_for(thing_name, key=None, session=None):
    """Read all the dweets for a dweeter
    """
    if key is not None:
        params = {'key': key}
    else:
        params = None
    return _request('get', '/get/dweets/for/{0}'.format(thing_name), params=params, session=session)


def remove_lock(lock, key, session=None):
    """Remove a lock (no matter what it's connected to).
    """
    return _request('get', '/remove/lock/{0}'.format(lock), params={'key': key}, session=session)


def lock(thing_name, lock, key, session=None):
    """Lock a thing (prevents unauthed dweets for the locked thing)
    """
    return _request('get', '/lock/{0}'.format(thing_name), params={'key': key, 'lock': lock}, session=session)


def unlock(thing_name, key, session=None):
    """Unlock a thing
    """
    return _request('get', '/unlock/{0}'.format(thing_name), params={'key': key}, session=session)


def set_alert(thing_name, who, condition, key, session=None):
    """Set an alert on a thing with the given condition
    """
    return _request('get', '/alert/{0}/when/{1}/{2}'.format(
        ','.join(who),
        thing_name,
        quote(condition),
    ), params={'key': key}, session=session)


def get_alert(thing_name, key, session=None):
    """Set an alert on a thing with the given condition
    """
    return _request('get', '/get/alert/for/{0}'.format(thing_name), params={'key': key}, session=session)


def remove_alert(thing_name, key, session=None):
    """Remove an alert for the given thing
    """
    return _request('get', '/remove/alert/for/{0}'.format(thing_name), params={'key': key}, session=session)


def get_stored_dweets_for(thing_name, key, date, hour='', response_type='', session=None):
    """Get stored dweets for the given thing during the given time.
    """
    return _request('get', '/get/stored/dweets/for/{0}'.format(thing_name),
                    params={'key': key, 'date': date, 'hour': hour, 'responseType': response_type}, session=session)



import setuptools

import dweet2ser

# except ImportError:
#    from distutils.core import setup

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='dweet2ser',
    version=dweet2ser.__version__,
    author='Zach Henry',
    author_email='zhenry9@gmail.com',
    description='A dweet.io <-> serial interface',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/zhenry9/dweet2ser',
    download_url='https://github.com/zhenry9/dweet2ser/archive/0.1.0.tar.gz',
    packages=setuptools.find_packages(exclude=('tests',)),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
    ],
    install_requires=[
        # 'dweepy',  packaging modified dweepy source with dweet2ser_conductor for now
        'requests >= 2, < 3',
        'pySerial',
        'colorama',
        'termcolor',
        # my patched dweepy repo, no longer needed
        # 'dweepy @ https://github.com/zhenry9/dweepy/archive/master.tar.gz',
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'dweet2ser = dweet2ser.__main__:main',
        ]
    },
    zip_safe=False,
)



import datetime
import os
import socket
from threading import Lock

from colorama import init as colorama_init, Fore, Style

colorama_init()

s_print_lock = Lock()


def _get_default_config_file():
    home_path = os.path.expanduser('~')
    # check if we're on linux running as superuser, then choose more appropriate directory
    if home_path == "/root":
        home_path = "/etc"
    else:
        home_path = os.path.join(home_path, '.config')
    file_name = "config.ini"
    return os.path.join(home_path, 'dweet2ser', file_name)


DEFAULT_CONFIG_FILE = _get_default_config_file()

# defaults when writing new configuration
CONFIG_DEFAULTS = {"type": "",
                   "location": "",
                   "port": "",
                   "thing_name": "dweet2ser_default",
                   "key": "None",
                   "mute": "False"
                   }

sys_stamp = "[ " + Fore.LIGHTBLACK_EX + "sys" + Style.RESET_ALL + " ] "


def timestamp():
    return "[" + Fore.LIGHTBLACK_EX + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") + Style.RESET_ALL + "] "


def internet_connection(host="8.8.8.8", port=53, timeout=3):
    """
        Host: 8.8.8.8 (google-public-dns-a.google.com)
        OpenPort: 53/tcp
        Service: domain (DNS/TCP)
        """
    try:
        socket.setdefaulttimeout(timeout)
        socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect((host, port))
        return True
    except socket.error as ex:
        #  print(ex)
        return False


def s_print(*a, **b):
    """Thread safe print function"""
    with s_print_lock:
        print(*a, **b)


def s_input(*a):
    """Thread safe input function"""
    with s_print_lock:
        print(*a, end='')
    return input('')



import argparse
import sys
import time

from colorama import init as colorama_init
from termcolor import colored

from dweet2ser import remote_device, local_device, device_bus
from dweet2ser.settings import sys_stamp, s_print, s_input
from dweet2ser.setup_config import Dweet2serConfiguration

colorama_init()

BUS = device_bus.DeviceBus()
CFG = Dweet2serConfiguration(BUS)


def add_device():
    name = s_input("\nDevice Name: ")
    location = s_input("Location (1.local 2.remote): ")
    mode = s_input("Type (DCE/DTE): ").upper()
    mute = s_input("Mute device? (y/n): ")
    if mute.upper().strip() == "Y" or mute.upper().strip() == "YES":
        mute = True
    else:
        mute = False
    d = None

    if location == "1":
        port = s_input("Port: ")
        try:
            d = local_device.LocalDevice(port, mode, name, mute)
        except Exception as e:
            s_print(e)

    elif location == "2":
        thing_id = s_input("Thing ID: ")
        key = s_input("Thing Key: ")
        if key == "None" or key == "":
            key = None
        try:
            d = remote_device.RemoteDevice(thing_id, mode, key, name, mute)
        except Exception as e:
            s_print(e)
    else:
        s_print("Invalid input")
        return
    if d:
        BUS.add_device(d)
    return


def remove_device():
    BUS.print_status()
    device = s_input("\nDevice to remove: ")
    BUS.remove_device(device)
    return


def process_input(cmd, ):
    if cmd == "info":
        return BUS.print_status()
    if cmd == "threads":
        return BUS.print_threads()
    if cmd == "add":
        return add_device()
    if cmd == "remove":
        return remove_device()
    if cmd == "save":
        return CFG.save_current_to_file()
    else:
        # print command help
        s_print("\tType 'info' to display session info.\n"
                "\tType 'add' to add a device.\n"
                "\tType 'remove' to remove a device.\n"
                "\tType 'save' to save the current configuration as default."
                )
        return


def idle():
    while True:
        time.sleep(1)


def main():
    arg_parser = argparse.ArgumentParser(description="An interface for networking R232 devices using dweet.io.")
    options = arg_parser.add_mutually_exclusive_group()
    options.add_argument("--file", type=str, help="Specify config file to use, overriding defaults.")
    options.add_argument("--empty", action='store_true', help="Start dweet2ser with no devices. "
                                                              "Useful for fixing broken config files "
                                                              "or creating new ones.")
    options.add_argument("--override", metavar=('MODE', 'PORT', 'THING_NAME'), action="store", type=str, nargs=3,
                         help="Setup a basic connection with command line arguments."
                              "\ne.g. --override DCE /dev/ttyUSB0 dweet2ser_default.")
    args = arg_parser.parse_args()

    s_print("\t\t*************************************************")
    s_print("\t\t**               " + colored("Dweet", "cyan") + " to " +
            colored("Serial", "red") + "               **")
    s_print("\t\t**                by Zach Henry                **")
    s_print("\t\t*************************************************")

    if args.override:
        if args.override[0].upper() == "DTE":
            other_mode = "DCE"
        else:
            other_mode = "DTE"
        try:
            local = local_device.LocalDevice(args.override[1], args.override[0])
            remote = remote_device.RemoteDevice(args.override[2], other_mode)
            BUS.add_device(local)
            BUS.add_device(remote)
        except Exception as e:
            s_print(f"{sys_stamp}Override failed: {e}")

    elif args.file:
        CFG.add_devices_from_file(args.file)

    elif not args.empty:
        CFG.add_devices_from_file()

    while True:
        cmd = ''
        try:
            time.sleep(.0001)
            cmd = s_input("\nType 'exit' to exit or ENTER for help.\n")
        except EOFError:  # if ran as a daemon, make sure we don't reach EOF prematurely
            idle()
        if cmd == 'exit':
            break
        process_input(cmd)


if __name__ == "__main__":
    sys.exit(main())



import datetime
import threading
import time

import requests
from termcolor import colored
from urllib3.exceptions import ProtocolError

from dweet2ser import dweepy
from dweet2ser.settings import timestamp, internet_connection, s_print


class DeadConnectionError(Exception):
    pass


class RemoteDevice(object):
    """
    Implementation of a serial device remotely connected with dweet.io
    """
    def __init__(self, thing_id, mode, thing_key=None, name="Remote Device", mute=False):
        self.name = name
        self.type = "dweet"
        self.type_color = "cyan"
        self.thing_id = thing_id
        self.locked = False
        self.thing_key = thing_key
        self.mute = mute
        if thing_key is not None:
            self.locked = True
        self.mode = mode

        if mode == 'DCE':  # if this is a DCE (device), set the dweet.io keywords appropriately
            self.write_kw = "from_pc"
            self.read_kw = "from_device"
        else:  # if this is a DTE (computer), the dweet.io keywords should be reversed.
            self.write_kw = "from_device"
            self.read_kw = "from_pc"

        self._session = requests.Session()
        self._last_message = ''
        self._kill_signal = "kill"
        self._message_queue = []
        self._started_on_day = datetime.datetime.utcnow().strftime("%Y-%m-%d")
        self.exc = False
        self.listening = False

    def write(self, message):
        """
        Tries to send a dweet message to the remote device. If there's no internet connection it saves
        the message to a queue.
        """
        if internet_connection():
            # check for a connection before trying to send to dweet
            self._send_dweet({self.write_kw: message})
            message_decoded = bytes.fromhex(message).decode('latin-1').rstrip()
            s_print(f"{timestamp()}{colored(self.type.capitalize(), self.type_color)} sent to {self.name}: "
                    f"{message_decoded}")
        else:
            # if there's no connection save the message to resend on reconnect
            s_print(f"{timestamp()}No connection to {self.name}. Saving message to queue.")
            self._message_queue.append(message)
            self.exc = True

    def send_message_queue(self):
        """
        Attempt to send queued messages.
        """
        while len(self._message_queue) > 0 and internet_connection():
            s_print(f"{timestamp()}Sending queued messages.")
            self.write(self._message_queue.pop(0))
            time.sleep(1.2)  # avoid exceeding dweet.io's 1s rate limit

    def _send_dweet(self, content):
        try:
            dweepy.dweet_for(self.thing_id, content, key=self.thing_key, session=self._session)

        except dweepy.DweepyError as e:
            s_print(timestamp() + str(e))
            s_print(f"{timestamp()}Trying again...")
            time.sleep(1.5)
            return self._send_dweet(content)

        except (ConnectionError, ProtocolError, OSError) as e:
            s_print(timestamp() + str(e))
            s_print(f"{timestamp()}Connection to {self.name} lost.")
            self.exc = True
            return

    def restart_session(self):
        """
        Gets a new Session from requests.
        """
        self._session = requests.Session()

    def listen(self):
        """
        Calls the listen/for/dweets/from function of dweet.io and yields messages from the chunked
        HTTP response.
        """
        self.listening = True
        # Start a thread to send a message to dweet every 45 seconds.
        # This is necessary because dweet.io closes the listen response after 60s of inactivity.
        keepalive_thread = threading.Thread(target=self._keepalive)
        keepalive_thread.daemon = True
        keepalive_thread.start()

        for message in self._listen_for_dweets():
            yield message

    def _listen_for_dweets(self):
        """ makes a call to dweepy to start a listening stream. error handling needs work
        """
        while internet_connection() and self.listening:
            try:
                for dweet in dweepy.listen_for_dweets_from(self.thing_id, key=self.thing_key,
                                                           timeout=90000, session=self._session):
                    content = dweet["content"]
                    self._last_message = dweet
                    if self.read_kw in content:
                        message = content[self.read_kw]
                        if message == self._kill_signal:
                            s_print(f"{timestamp()}Listen stream for {self.name} closed.")
                            return
                        yield message

            # if you get an error because dweet closed the connection, open it again.
            except (ConnectionError, ProtocolError, OSError) as e:
                s_print(timestamp() + str(e))
                s_print(f"{timestamp()}Connection closed by dweet, restarting.")
                self.restart_session()
                yield self.get_last_message()
        s_print(f"{timestamp()}Listen stream for {self.name} closed.")
        self.exc = True
        return

    def kill_listen_stream(self):
        """
        Attempts to manually end the chunked HTTP listening stream.
        """
        self._send_dweet({self.read_kw: self._kill_signal})
        self.listening = False

    def _keepalive(self):
        """ dweet.io seems to close the connection after 60 seconds of inactivity.
            This sends a dummy payload every 45s to avoid that.
        """
        while self.listening and not self.exc:
            time.sleep(45)
            self._send_dweet({"keepalive": 1})

    def get_last_message(self):
        """
        Tries to recover the last message sent to dweet.
        """
        message = ''
        if internet_connection():
            dweet = dweepy.get_latest_dweet_for(self.thing_id, key=self.thing_key, session=self._session)
            content = dweet[0]["content"]
            if self.read_kw in content:
                message = content[self.read_kw]
        return message

    @staticmethod
    def _get_dweet_time(dweet):
        """
        For future dweet recovery implementation.
        """
        created = dweet["created"].replace("T", " ").replace("Z", "")
        dweet_time = datetime.datetime.fromisoformat(created)
        return dweet_time

# TODO Create function to retrieve dweets from storage after a connection loss.
#   Basic idea: store time of last_message_received from listen.
#   On connection restore, pull all dweets from today and yield all where "created" > last_message_received
